# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from .common import PrologueError, Line
from .context import Context
from .directives import register_prime_directives
from .directives.base import Directive
from .directives.common import DirectiveWrap
from .registry import Registry

class Prologue(object):
    """ Top-level of the preprocessor """

    def __init__(
        self,
        comment         ="#",
        delimiter       ="#",
        shared_delimiter=False,
        implicit_sub    =True,
        explicit_style  =("$(", ")"),
        allow_redefine  =False,
        register_prime  =True,
    ):
        """ Initialise the preprocessor.

        Args:
            comment         : The single line comment sequence (default: '#')
            delimiter       : Specify the sequence to use as the delimiter
                              (default: '#')
            shared_delimiter: Delimiter is also used to signify a comment
                              (default: False)
            implicit_sub    : Allow implicit substitutions (default: True)
            explit_style    : Style used for identifying explicit substitutions,
                              this should be a tuple of the prefix and suffix of
                              the format (default: '$(', ')')
            allow_redefine  : Allow values to be defined multiple times (by
                              default raises a PrologueError, default: False)
            register_prime  : Register the prime directives (default: True)
        """
        # Sanity checks
        if not isinstance(comment, str):
            raise PrologueError(f"Comment sequence must be a string: {comment}")
        if not isinstance(delimiter, str):
            raise PrologueError(f"Delimiter sequence must be a string: {delimiter}")
        if shared_delimiter not in (True, False):
            raise PrologueError(f"Shared delimiter must be True or False: {shared_delimiter}")
        if implicit_sub not in (True, False):
            raise PrologueError(f"Implicit substitution must be True or False: {implicit_sub}")
        if not isinstance(explicit_style, tuple):
            raise PrologueError(f"Explicit style must be a tuple: {explicit_style}")
        if allow_redefine not in (True, False):
            raise PrologueError(f"Allow redefinition must be True or False: {allow_redefine}")
        if register_prime not in (True, False):
            raise PrologueError(f"Register prime must be True or False: {register_prime}")
        # Store attributes
        self.comment          = comment
        self.delimiter        = delimiter
        self.shared_delimiter = shared_delimiter
        self.implicit_sub     = implicit_sub
        self.explicit_style   = explicit_style
        self.allow_redefine   = allow_redefine
        # Create empty message handling callbacks
        self.callback_debug   = None
        self.callback_info    = None
        self.callback_warning = None
        self.callback_error   = None
        # Create a registry instance
        self.registry = Registry(self)
        # Create a store for directives
        self.directives = {}
        if register_prime: register_prime_directives(self)
        # Setup blank loop for converting from output line to input file and line
        self.lookup = None

    # ==========================================================================
    # Property Setters/Getters
    # ==========================================================================

    @property
    def delimiter(self): return self.__delimiter

    @delimiter.setter
    def delimiter(self, val):
        # Check delimiter
        if len(val.strip()) == 0:
            raise PrologueError("Delimiter should be at least one character")
        elif len(val.replace(" ", "")) != len(val):
            raise PrologueError("Delimiter should not contain whitespace")
        # Set delimiter
        self.__delimiter = val

    @property
    def shared_delimiter(self): return self.__shared_delimiter

    @shared_delimiter.setter
    def shared_delimiter(self, val):
        # Check value is True or False
        if not isinstance(val, bool):
            raise PrologueError("Shared delimiter should be True or False")
        # Set value
        self.__shared_delimiter = val

    # ==========================================================================
    # Message Handlers
    # ==========================================================================

    def debug_message(self, message, **kwargs):
        """ Handle debug level messages.

        Args:
            message: The message to print
            kwargs : Attributes related to the message
        """
        # If a callback is configured, delegate message to it
        if self.callback_debug:
            self.callback_debug(message, **kwargs)
        # Otherwise print it out
        else:
            print(f"[PROLOGUE:DEBUG] {message}")

    def info_message(self, message, **kwargs):
        """ Handle info level messages.

        Args:
            message: The message to print
            kwargs : Attributes related to the message
        """
        # If a callback is configured, delegate message to it
        if self.callback_info:
            self.callback_info(message, **kwargs)
        # Otherwise print it out
        else:
            print(f"[PROLOGUE:INFO] {message}")

    def warning_message(self, message, **kwargs):
        """ Handle warning level messages.

        Args:
            message: The message to print
            kwargs : Attributes related to the message
        """
        # If a callback is configured, delegate message to it
        if self.callback_warning:
            self.callback_warning(message, **kwargs)
        # Otherwise print it out
        else:
            print(f"[PROLOGUE:WARN] {message}")

    def error_message(self, message, **kwargs):
        """ Handle error level messages.

        Args:
            message: The message to print
            kwargs : Attributes related to the message
        """
        # If a callback is configured, delegate message to it
        if self.callback_error:
            self.callback_error(message, **kwargs)
        # Otherwise print it out
        else:
            raise PrologueError(message)

    # ==========================================================================
    # Registry Passthroughs
    # ==========================================================================

    def add_file(self, path):
        """ Add a file to the registry.

        Args:
            path: Path to add
        """
        self.registry.add_file(path)

    def add_folder(self, path, search_for=None, recursive=False):
        """ Add a folder to the registry, see Registry.add_folder for more info.

        Args:
            path      : Path to the root folder to add
            search_for: Provide a file extension to search for
            recursive : Whether to search recursively in this folder
        """
        self.registry.add_folder(path, search_for=search_for, recursive=recursive)

    # ==========================================================================
    # Directives
    # ==========================================================================

    def register_directive(self, dirx):
        """ Register a directive for handling lines or blocks.

        Args:
            dirx: Decorated function that will act as the directive
        """
        # Check that the directive is correctly decorated
        if not isinstance(dirx, DirectiveWrap):
            raise PrologueError("Directive type is not known, is it decorated?")
        # Check all tags
        for tag in dirx.tags:
            # Check if the tag collides with an existing directive
            if tag.lower() in self.directives:
                raise PrologueError(f"Directive already registered for tag '{tag}'")
        # Register the directive
        for tag in dirx.tags: self.directives[tag.lower()] = dirx

    def deregister_directive(self, tag):
        """ Remove a previously registered directive.

        Args:
            tag: The tag of the directive
        """
        if tag.lower() not in self.directives:
            raise PrologueError(f"No directive registered for tag '{tag}'")
        for dirx_tag in self.directives[tag.lower()].tags:
            del self.directives[dirx_tag]

    def list_directives(self):
        """ Return all of the registered directives.

        Returns: A list of directives """
        return list(set(self.directives.values()))

    def has_directive(self, tag):
        """ Test if a directive has been registered for a given tag.

        Args:
            tag: The tag to check for

        Returns: True if the tag is known, False if not
        """
        return (tag.lower() in self.directives)

    def get_directive(self, tag):
        """
        Return a directive for a particular tag. If the delimiter is not shared
        and the tag isn't known, an error will be raised. If the delimiter is
        shared, then None will be returned.

        Args:
            tag: The tag of the directive

        Returns: Directive function if known, otherwise None
        """
        # Check if a directive exists for this tag
        if tag.lower() not in self.directives:
            if self.shared_delimiter:
                return None
            else:
                raise PrologueError(f"No directive known for tag '{tag}'")
        # Return the directive
        return self.directives[tag.lower()]

    # ==========================================================================
    # Evaluation
    # ==========================================================================

    def evaluate(self, filename, defines=None, callback=None):
        """ Iterable evaluation function which returns fully preprocessed stream

        Args:
            filename: The file to evaluate
            defines : Predefined values to initialise the context (default: None)
            callback: Optional callback provided to all directives to expose
                      state during parse to an external routine (default: None)

        Yields: Stream of fully preprocessed lines
        """
        # Create a context object used for tracking variables and parse state
        context = Context(
            self,
            implicit_sub  =self.implicit_sub,
            explicit_style=self.explicit_style,
            allow_redefine=self.allow_redefine,
            initial_state =defines,
        )
        # Use inner evaluation routine to get each line one at a time
        self.lookup = []
        for line in self.evaluate_inner(filename, context, callback=callback):
            final = context.substitute(line)
            self.lookup.append((final.file, final.number))
            yield str(final)

    def evaluate_inner(self, filename, context, callback=None):
        """
        Inner evaluation routine - this performs all construction and evaluation
        except variable substitution, which is performed by 'evaluate'.

        Args:
            filename: The file to evaluate
            context : A context object
            callback: Optional callback provided to all directives to expose
                      state during parse to an external routine (default: None)

        Yields: Stream of preprocessed lines, prior to substitution
        """
        # Find the top-level file
        r_file = self.registry.resolve(filename)
        if not r_file: raise PrologueError(f"Failed to find file {filename}")
        # Sanity check a context object was provided
        if not isinstance(context, Context):
            raise PrologueError(f"An invalid context was provided: {context}")
        # Stop infinite recursion by checking if this file is already in the stack
        if r_file in context.stack:
            raise PrologueError(
                f"Detected infinite recursion when including file '{filename}' "
                f"- file stack: {', '.join([x.filename for x in context.stack])}"
            )
        # Create regular expressions for recognising directives
        re_anchored = re.compile(
            r"^[\s]*[" + self.delimiter + r"]([a-z0-9_]+)(.*?)$", flags=re.IGNORECASE,
        )
        re_floating = re.compile(
            r"^([^" + self.comment + "]*)[" + self.delimiter + r"]([a-z0-9_]+)(.*?)$", flags=re.IGNORECASE,
        )
        # Push the current file into the stack
        context.stack_push(r_file)
        # Start parsing
        active      = None
        accumulated = None
        for idx, line in enumerate(r_file.contents):
            # Catch any exceptions so that they can be marked with file and line
            try:
                # If comment and delimiter are different, remove everything after comment
                if self.comment != self.delimiter:
                    line = line.split(self.comment)[0]
                # Handle line continuation
                if line and line[-1] == "\\":
                    accumulated = (accumulated + line[:-1]) if accumulated else line[:-1]
                    continue
                elif accumulated:
                    line        = accumulated + line
                    accumulated = None
                # Test if the line matches an anchored directive
                anchored = re_anchored.match(line)
                if anchored:
                    tag, arguments = anchored.groups()
                    arguments      = arguments.strip()
                    tag            = tag.lower()
                    d_wrap         = self.get_directive(tag)
                    if arguments.endswith(":"): arguments = arguments[:-1]
                    if d_wrap and d_wrap.is_line:
                        l_dir = d_wrap.directive(
                            active, src_file=r_file, src_line=(idx + 1),
                            callback=callback,
                        )
                        l_dir.invoke(tag, arguments.strip())
                        if   active      : active.append(l_dir)
                        elif l_dir.yields: yield from l_dir.evaluate(context)
                        else             : l_dir.evaluate(context)
                        # Move on to the next line
                        continue
                    elif d_wrap and d_wrap.is_block:
                        # Call the directive
                        if d_wrap.is_opening(tag):
                            block   = d_wrap.directive(
                                active, src_file=r_file, src_line=(idx + 1),
                                callback=callback,
                            )
                            block.open(tag, arguments)
                            # If a block is already open, append to it
                            if active: active.append(block)
                            # Track currently active block
                            active = block
                        elif d_wrap.is_transition(tag):
                            if d_wrap.directive != type(active):
                                raise PrologueError(f"Transition tag '{tag}' was not expected")
                            active.transition(tag, arguments)
                        elif d_wrap.is_closing(tag):
                            if d_wrap.directive != type(active):
                                raise PrologueError(f"Closing tag '{tag}' was not expected")
                            active.close(tag, arguments)
                            # If there is no parent, this is the root
                            if not active.parent:
                                if active.yields:
                                    yield from active.evaluate(context.fork())
                                else:
                                    active.evaluate(context.fork())
                            # Pop the stack
                            active = active.parent
                        # Move on to the next line
                        continue
                # Test if the line matches a floating directive
                floating = re_floating.match(line)
                if floating != None:
                    prior, tag, arguments = floating.groups()
                    tag                   = tag.lower()
                    arguments             = arguments.strip()
                    d_wrap                = self.get_directive(tag)
                    if d_wrap and d_wrap.is_block:
                        raise PrologueError(
                            f"The directive '{tag}' can only be used with an "
                            f"anchored delimiter as it is a block directive"
                        )
                    elif d_wrap:
                        if arguments.endswith(":"): arguments = arguments[:-1]
                        # Yield the text before the directive
                        yield line.encase(prior.rstrip())
                        # Yield the contents returned from the directive
                        l_dir = d_wrap.directive(
                            active, src_file=r_file, src_line=(idx + 1),
                            callback=callback,
                        )
                        l_dir.invoke(tag, arguments.strip())
                        if   active      : active.append(l_dir)
                        elif l_dir.yields: yield from l_dir.evaluate(context)
                        else             : l_dir.evaluate(context)
                        continue
                # Otherwise, this is just a line!
                if active: active.append(line)
                else     : yield line
            # Catch all exceptions and add file and line
            except PrologueError as e:
                raise e
            except Exception as e:
                snippet = "\n".join(r_file.snippet(idx+1))
                raise PrologueError(
                    f"Caught {type(e).__name__} when evaluating {r_file.path}:"
                    f"{idx+1}; {e}" + "\n" + snippet
                ) from e
        # Check for trailing directives
        if active:
            dir_stack = [x for x in active.stack if isinstance(x, Directive)]
            src_file, src_line = active.source
            snippet = "\n".join(src_file.snippet(src_line))
            raise PrologueError(
                f"Unmatched {type(active).__name__} block directive in "
                f"{src_file.path}:{src_line}:" + "\n" + snippet
            )
        # Pop the file being parsed from the stack
        if context.stack_pop() != r_file:
            raise PrologueError("File stack has been corrupted")

    # ==========================================================================
    # Lookup
    # ==========================================================================

    def resolve(self, line, before=2, after=2):
        """
        Use the line lookup to resolve which input file and line number produced
        each line of the output.

        Args:
            line  : Line number (from 1 to number of output lines, not an index)
            before: How many lines to include in the snippet before the target
            after : How many lines to include in the snippet after the target

        Returns: Tuple of input RegistryFile and the line number
        """
        # Sanity checks
        if not self.lookup:
            raise PrologueError(
                "Lookup does not yet exist - have you called 'evaluate'?"
            )
        elif not isinstance(line, int):
            raise PrologueError(f"Line number must be an integer - not '{line}'")
        elif line < 1 or line > len(self.lookup):
            raise PrologueError(
                f"Line {line} is out of valid range 1-{len(self.lookup)}"
            )
        # Use lookup to get the file and line number
        r_file, line_no = self.lookup[line-1]
        # Return the input file and line number
        return r_file, line_no, r_file.snippet(line_no, before, after)



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import BlockDirective
from ..block import Block
from .common import directive
from ..common import PrologueError

@directive(
    opening   =["if", "ifdef", "ifndef"],
    transition=["elif", "else"],
    closing   =["endif"]
)
class Conditional(BlockDirective):
    """ Conditional directive supporting IF/ELIF/ELSE/ENDIF sections """

    def __init__(self, parent, src_file=None, src_line=0, callback=None):
        super().__init__(
            parent, src_file=src_file, src_line=src_line, callback=callback
        )
        self.if_section    = None
        self.elif_sections = []
        self.else_section  = None

    def open(self, tag, arguments):
        """ Open the conditional block with an 'if' clause.

        Args:
            tag      : The tag that opens the directive
            arguments: Argument string following the directive
        """
        # Sanity checks
        if tag not in ("if", "ifdef", "ifndef"):
            raise PrologueError(f"Conditional opening invoked with '{tag}'")
        # Now open the tag
        super().open(tag, arguments)
        # Record the section
        self.if_section = tag, arguments, Block(self)

    def transition(self, tag, arguments):
        """ Transition between different sections with ELIF/ELSE.

        Args:
            tag      : The tag that opens the section
            arguments: Argument string following the directive
        """
        # Sanity checks
        if tag not in ["elif", "else"]:
            raise PrologueError(f"Conditional transition invoked with '{tag}'")
        elif self.else_section:
            raise PrologueError(f"Transition '{tag}' detected after 'else' clause")
        # Now perform transition
        super().transition(tag, arguments)
        # Register the tag
        if tag == "elif":
            self.elif_sections.append((tag, arguments, Block(self)))
        else:
            self.else_section = tag, arguments, Block(self)

    def close(self, tag, arguments):
        """ Close the directive block with ENDIF.

        Args:
            tag      : The tag that opens the section
            arguments: Argument string following the directive
        """
        # Sanity checks
        if tag != "endif":
            raise PrologueError(f"Conditional close invoked with '{tag}'")
        # Now close tag
        super().close(tag, arguments)

    def append(self, entry):
        """ Append a new line or nested block to the active section.

        Args:
            entry: The line or block to append.
        """
        # Sanity checks
        if not self.opened:
            raise PrologueError("Trying to append a line to an unopened conditional")
        elif self.closed:
            raise PrologueError("Trying to append a line to a closed conditional")
        # Append to the latest open block
        if   self.else_section          : self.else_section[2].append(entry)
        elif len(self.elif_sections) > 0: self.elif_sections[-1][2].append(entry)
        else                            : self.if_section[2].append(entry)

    def evaluate(self, context):
        """ Selects the correct block to evaluate based upon conditions.

        Args:
            context: The context object at the point of evaluation

        Yields: A line of text at a time
        """
        # Put sections in the correct order to evaluate
        sections = [self.if_section, *self.elif_sections]
        if self.else_section: sections.append(self.else_section)
        # Check which section is active
        for tag, cond, block in sections:
            # Replace alternative syntaxes for Python operators
            cond = cond.replace("&&", " and ").replace("||", " or ")
            # Evaluate the conditional
            if (
                (tag == "ifdef"  and     context.has_define(cond)) or
                (tag == "ifndef" and not context.has_define(cond)) or
                (tag not in ("ifdef", "ifndef") and context.evaluate(cond)) or
                (tag == "else")
            ):
                yield from block.evaluate(context)
                context.join()
                break



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntEnum
from pathlib import Path

from .common import PrologueError, Line

class RegistryFile(object):
    """ Holds a file in the registry """

    def __init__(self, path):
        """ Initialise the registry file

        Args:
            path: Path to the file
        """
        # Ensure that the path is a Path instance
        self.path = Path(path) if not isinstance(path, Path) else path
        # Check the path exists
        if not self.path.exists():
            raise PrologueError(
                f"File does not exist at path {self.path}"
            )
        # Check that the path is a file
        if not self.path.is_file():
            raise PrologueError(
                f"Path provided is not a file {self.path}"
            )

    @property
    def filename(self): return self.path.parts[-1]

    @property
    def contents(self):
        """ Provide an iterable that reads a line at a time from the file """
        with open(self.path, "r") as fh:
            for idx, line in enumerate(fh.readlines()):
                yield Line(line.rstrip(), self, idx + 1)

    def snippet(self, line, before=2, after=2):
        """ Generate a snippet of the original file.

        Args:
            line  : Line number to extract
            before: Number of lines before the line to include
            after : Number of lines after the line to include

        Returns: List of lines forming snippet
        """
        snippet = []
        for s_line in self.contents:
            # If this is before the snip point, ignore it
            if s_line.number < (line - before): continue
            # If this is after the snip point, break out
            if s_line.number > (line + after): break
            # Otherwise, build up the snippet
            snippet.append("%4i %s %s" % (
                s_line.number, (">>" if (s_line.number == line) else "  "),
                str(s_line)
            ))
        return snippet

class RegistryFolder(object):
    """ Holds a folder in the registry """

    def __init__(self, path):
        """ Initialise the registry folder

        Args:
            path: Path to the folder
        """
        # Ensure that the path is a Path instance
        self.path = Path(path) if not isinstance(path, Path) else path
        # Check the path exists
        if not self.path.exists():
            raise PrologueError(
                f"Folder does not exist at path {self.path}"
            )
        # Check that the path is a directory
        if not self.path.is_dir():
            raise PrologueError(
                f"Path provided is not a folder {self.path}"
            )

    @property
    def folder(self): return self.path.parts[-1]

    def resolve(self, path):
        """ Resolve a relative path within this folder to get a file.

        Args:
            path: The path to resolve
        """
        # Ensure this is a Path object
        path = Path(path) if not isinstance(path, Path) else path
        # Check if this is absolute
        if path.is_absolute():
            raise PrologueError(f"Attempted to resolve absolute path {path}")
        # Check how many segments are left
        if len(path.parts) > 1:
            return RegistryFolder(self.path / path.parts[0]).resolve(
                Path(*path.parts[1:])
            )
        else:
            return RegistryFile(self.path / path.parts[0])

class Registry(object):
    """ Keeps track of all files available to IMPORT and INCLUDE """

    def __init__(self, pro, flat=False):
        """ Initialise the registry

        Args:
            pro : Pointer to the Prologue instance
            flat: Flatten all hierarchy of folders added to the registry.
        """
        self.__pro     = pro
        self.__flat    = flat
        self.__entries = {}

    def insert_entry(self, entry):
        """ Insert an entry into the registry.

        Args:
            entry: Either a RegistryFile or RegistryFolder instance
        """
        # Check the entry type
        if type(entry) not in [RegistryFile, RegistryFolder]:
            raise PrologueError(
                f"Entry is not a recognised type {type(entry).__name__}"
            )
        # Check that the entry doesn't collide
        entry_name = (
            entry.filename if isinstance(entry, RegistryFile) else entry.folder
        )
        if entry_name in self.__entries:
            raise PrologueError(
                f"Entry already exists in registry with name {entry_name}"
            )
        # Insert the entry
        self.__pro.debug_message(
            f"Adding entry '{entry_name}' to registry: {entry.path}"
        )
        self.__entries[entry_name] = entry

    def add_file(self, path):
        """ Add a specific file to the registry.

        Args:
            path: Path to the file to add
        """
        self.insert_entry(RegistryFile(path))

    def add_folder(self, path, search_for=None, recursive=False):
        """
        Add a new folder to the registry. If 'search_for' is provided, then all
        matching files within the folder will be added to the registry root,
        optionally if 'recursive' is provided the whole tree beneath the folder
        will be traversed. If only 'recursive' is provided, all folders within
        the top level folder will be added. If neither argument is provided,
        only the top-level folder will be added.

        Args:
            path      : Path to the root folder to add
            search_for: Provide a file extension to search for
            recursive : Whether to search recursively in this folder
        """
        self.__pro.debug_message(
            "Adding folder " + ("" if recursive else "non-") + "recursively to "
            f"registry, searching for {search_for}: {path}"
        )
        # Wrap the folder
        r_folder = RegistryFolder(path)
        # If neither argument provided, register directly
        if not self.__flat and not search_for and not recursive:
            self.insert_entry(r_folder)
        # If only recursive is provided, register all subfolders
        elif not self.__flat and not search_for and recursive:
            for subfolder in r_folder.path.rglob("**/"):
                self.insert_entry(RegistryFolder(subfolder))
        # If search_for is provided, lookup files
        elif self.__flat or isinstance(search_for, str):
            if not search_for: search_for = ""
            for subpath in (
                r_folder.path.rglob(f"**/*{search_for}")
                if (self.__flat or recursive) else
                r_folder.path.glob(f"*{search_for}")
            ):
                if subpath.is_dir(): continue
                self.insert_entry(RegistryFile(subpath))
        # Unknown scenario
        else:
            raise PrologueError(f"Unexpected error adding folder {path}")

    def resolve(self, path):
        """ Resolve a path within the registry to get a file.

        Args:
            path: The path to resolve
        """
        # Ensure this is a Path instance
        path = Path(str(path).strip())
        # Check if the path is absolute
        if path.is_absolute(): return RegistryFile(path)
        # Otherwise, attempt to resolve
        if path.parts[0] not in self.__entries:
            raise PrologueError(f"No registry entry found for '{path}'")
        # Lookup the entry
        entry = self.__entries[path.parts[0]]
        # Sanity check
        if len(path.parts) > 1:
            if isinstance(entry, RegistryFile):
                raise PrologueError(f"Only a file is registered for path {path.parts[0]}")
            return entry.resolve(Path(*path.parts[1:]))
        elif isinstance(entry, RegistryFolder):
            raise PrologueError(f"Failed to resolve {path} to a file")
        else:
            return entry



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class PrologueError(Exception):
    """ Custom exception class - raised when Prologue detects processing errors """
    pass

class Line(str):
    """ Carries a line from a file, along with the line number and file pointer """

    def __new__(cls, line, file, number):
        return super().__new__(cls, line)

    def __init__(self, line, file, number):
        """ Initialise the line object.

        Args:
            line  : The string representing the line
            file  : Pointer to the file that contains this line
            number: The line number (minimum value of 1) within the file
        """
        super().__init__()
        assert isinstance(number, int) and number >= 1
        self.file   = file
        self.number = number

    def __repr__(self):
        return f"{self.file}@{self.number}: {self.__str__()}"

    def encase(self, substring):
        """ Encase a substring with the same file pointer and line number.

        Args:
            substring: The substring to encase
        """
        return Line(str(substring), self.file, self.number)

    def __getitem__(self, item):
        """ Custom handler for accessing single character or range in a line.

        Args:
            item: Character index or range to retrieve
        """
        return self.encase(super().__getitem__(item))

    def split(self, *args, **kwargs):
        return [self.encase(x) for x in super().split(*args, **kwargs)]

    def strip(self, *args, **kwargs):
        return self.encase(super().strip(*args, **kwargs))

    def __add__(self, other):
        return self.encase(super().__add__(other))



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import re
import shlex
import sys

# Support AST unparsing across multiple Python versions
# NOTE: Before Python 3.9, unparse was not a native function
if sys.version_info >= (3, 9):
    from ast import unparse
else:
    from astunparse import unparse

from .common import PrologueError, Line
from .registry import RegistryFile

class Context(object):
    """ Keeps track of the parser's context """

    def __init__(
        self,
        pro,
        parent        =None,
        implicit_sub  =True,
        explicit_style=("$(", ")"),
        allow_redefine=False,
        initial_state =None,
    ):
        """ Initialise the context

        Args:
            pro           : Pointer to the root Prologue instance
            parent        : Pointer to the parent Context object prior to fork
            implicit_sub  : Whether to allow implicit substitutions (default: True)
            explicit_style: Tuple of strings that define the explicit style
                            (default: ('$(', ')'))
            allow_redefine: Allow values to be defined multiple times (default:
                            False, so a PrologueError will be raised on redefine)
            initial_state : Dictionary of values to provide an initial state
                            (default: None)
        """
        self.pro            = pro
        self.parent         = parent
        self.implicit_sub   = implicit_sub
        self.explicit_style = explicit_style
        self.allow_redefine = allow_redefine
        self.__defines      = {}
        self.__removed      = []
        self.__stack        = []
        self.__trace        = []
        # Populate initial state
        if isinstance(initial_state, dict):
            for key, value in initial_state.items():
                if not isinstance(key, str):
                    raise PrologueError(
                        f"Keys must be strings in initial state, not: {key}"
                    )
                elif type(value) not in (str, int, float, bool):
                    raise PrologueError(
                        f"Values must be string, integer, float, or Boolean in "
                        f"the initial state, not: {value}"
                    )
                self.__defines[key] = value
        elif initial_state != None:
            raise PrologueError(
                f"Initial state must be a dictionary, not: {initial_state}"
            )
        # Define regular expression for variable substitution
        rgx_exp_str = r"(" + "".join(f"[{x}]" for x in self.explicit_style[0])
        if explicit_style[1]:
            rgx_exp_str += r".*?"
            rgx_exp_str += "".join(f"[{x}]" for x in self.explicit_style[1])
        else:
            rgx_exp_str += r"[a-z_][a-z0-9_]+"
        rgx_exp_str += r")"
        self.rgx_exp = re.compile(rgx_exp_str, flags=re.IGNORECASE)
        self.rgx_imp = re.compile(r"\b([a-z][a-z0-9_]{0,})\b", flags=re.IGNORECASE)

    @property
    def defines(self):
        """ Returns all defines across the full stack of context objects """
        return {
            **(self.parent.defines if self.parent else {}), **self.__defines
        }

    @property
    def root(self):
        """ Returns the root context object """
        return self.parent.root if self.parent else self

    @property
    def stack(self):
        """ Returns the root file stack object """
        return self.root.stack if self.parent else self.__stack

    @property
    def trace(self):
        """ Returns the root file trace object """
        return self.root.trace if self.parent else self.__trace

    # ==========================================================================
    # File Stack Management
    # ==========================================================================

    def stack_push(self, file):
        """ Push a new file onto the stack.

        Args:
            file: Instance of RegistryFile to push
        """
        # Sanity check this is the correct type
        if not isinstance(file, RegistryFile):
            raise PrologueError(
                f"Trying to push {file} to stack - must be a RegistryFile"
            )
        # Push to stack
        self.stack.append(file)
        # Also push file to the trace - it records order files were read
        self.trace.append(file)

    def stack_pop(self):
        """ Pop a file from the stack.

        Returns: Next instance of RegistryFile from the top of the stack.
        """
        # Sanity check
        if len(self.stack) == 0:
            raise PrologueError("Trying to pop file from empty stack")
        # Pop the top entry off the stack
        return self.stack.pop()

    def stack_top(self):
        """ Get the top RegistryFile instance from the stack (if any exist).

        Returns: RegistryFile instance if stack populated, else None.
        """
        # Return the top item off the stack
        return self.stack[-1] if len(self.stack) > 0 else None

    # ==========================================================================
    # Defined Constant Handling
    # ==========================================================================

    def set_define(self, key, value, check=True):
        """ Define a value for a key, checks whether the key clashes.

        Args:
            key  : The key of the define
            value: The value of the define
            check: Check whether an existing variable is about to be redefined
        """
        # Check the key is sane
        if " " in key or len(key) == 0:
            raise PrologueError(
                f"Key must not contain whitespace and must be at least one "
                f"character in length: '{key}'"
            )
        # Warn about collision
        if check:
            if key in self.defines and not key in self.__removed:
                msg = (
                    f"Value already defined for key '{key}' with value "
                    f"{self.get_define(key)}"
                )
                if self.allow_redefine:
                    self.pro.warning_message(msg, key=key, value=value)
                else:
                    raise PrologueError(msg)
        # If the value is a number, convert it
        if isinstance(value, str) and value.strip().isdigit(): value = int(value)
        # Store the define
        self.__defines[key] = value
        # Clear define name from 'removed' array, if present
        if key in self.__removed: self.__removed.remove(key)

    def clear_define(self, key):
        """
        Remove a defined value from the context based on a key, an exception is
        raised if the key is unknown.

        Args:
            key: The key of the defined value
        """
        if key not in self.defines:
            raise PrologueError(f"No value has been defined for key '{key}'")
        if key in self.__defines:
            del self.__defines[key]
        else:
            self.__removed.append(key)

    def has_define(self, key):
        """ Check if a variable has been defined.

        Args:
            key: The key of the defined value

        Returns: True if defined, False otherwise
        """
        # If removed at this level, immediately return False
        if key in self.__removed: return False
        # If defined at this level, immediately return True
        if key in self.__defines: return True
        # Check if defined by my parent instead?
        if self.parent: return self.parent.has_define(key)
        # Otherwise, it's not defined!
        return False

    def get_define(self, key):
        """ Return a value if the key is known, otherwise raise an exception.

        Args:
            key: The key of the defined value

        Returns: Value of the define
        """
        if key in self.__removed:
            raise PrologueError(f"No value has been defined for key '{key}'")
        elif key not in self.__defines:
            if self.parent:
                return self.parent.get_define(key)
            else:
                raise PrologueError(f"No value has been defined for key '{key}'")
        return self.defines[key]

    # ==========================================================================
    # Forking and Joining
    # ==========================================================================

    def fork(self):
        """
        Creates a new context object with tracking of parent instance. This
        allows variable state to propagate from the parent, but be overridden
        locally - which is vital in supporting nested block evaluation.

        Returns: Instance of Context
        """
        return Context(self.pro, parent=self, implicit_sub=self.implicit_sub)

    def join(self):
        """ Joins a child context object back into it's parent.

        Returns: Pointer to the parent context
        """
        if not self.parent:
            raise PrologueError("No parent configured for context object")
        for key in self.__removed:
            self.parent.clear_define(key)
        for key, value in self.__defines.items():
            self.parent.set_define(key, value, check=False)
        return self.parent

    # ==========================================================================
    # Expression Evaluation
    # ==========================================================================

    def flatten(self, expr, history=None):
        """ Flatten an expression by substituting for known variables.

        Args:
            expr   : The expression to flatten
            history: Tracks expressions to avoid deadlock

        Returns: String with each recognised variable substituted for its value
        """
        # Create a history if one doesn't already exist
        if not history: history = []
        history.append(expr)
        # Declare a AST node transformation to replace variables
        ctx      = self
        replaced = [0]
        class ReplaceVar(ast.NodeTransformer):
            def visit_Name(self, node):
                if ctx.has_define(node.id):
                    replaced[0] += 1
                    value = ctx.get_define(node.id)
                    if not isinstance(value, str):
                        return ast.Constant(value=value, kind=type(value))
                    else:
                        return ast.parse(value).body[0]
                else:
                    return ast.Constant(value=node.id, kind=type(node.id))
        # Iterate repeating variables (constants may reference other constants)
        result = expr
        while True:
            # If result is no longer a string, break out
            if not isinstance(result, str): break
            # If it's not a constant, then try to substitute
            replaced[0] = 0
            try:
                # Walk the AST substituting variables -> constants
                result = unparse(ReplaceVar().visit(ast.parse(result)))
            except TypeError:
                break
            if replaced[0] == 0: break
        return result

    def evaluate(self, expr):
        """ Flatten an expression, then evaluate it.

        Args:
            expr: The expression to evaluate

        Returns: Result of the expression
        """
        # First flatten out variable references
        flat = self.flatten(expr.strip())
        # Now evaluate (if we can)
        try:
            return eval(flat, { "__builtins__": None }, { })
        except Exception:
            return flat

    def substitute(self, line, implicit=None):
        """ Perform in-line substitutions for recognised variables.

        Args:
            line    : The line to perform substitution on
            implicit: Enable implicit substitutions

        Returns: Line with values substituted
        """
        # If implicit not given, then it defaults to instance version
        if implicit == None: implicit = self.implicit_sub
        # Pickup the input line's attributes
        f_file, f_line, line = line.file, line.number, str(line)
        # First look for explicit substitutions of the form '$(x)'
        exp_match = [x for x in self.rgx_exp.finditer(line)]
        final     = ""
        exp_match = [x for x in self.rgx_exp.finditer(line)]
        for match in exp_match[::-1]:
            m_expr  = match.groups()[0][
                len(self.explicit_style[0]):
                len(match.groups()[0])-len(self.explicit_style[1])
            ]
            sub_val = self.evaluate(m_expr)
            line    = (
                line[:match.span()[0]] + str(sub_val) + line[match.span()[1]:]
            )
        # Secondly look for implicit substitutions
        if implicit:
            imp_match = [x for x in self.rgx_imp.finditer(line)]
            # Substitute each identified variable
            for match in imp_match[::-1]:
                if self.has_define(match.groups()[0]):
                    line = (
                        line[:match.span()[0]] +
                        str(self.evaluate(match.groups()[0])) +
                        line[match.span()[1]:]
                    )
        # Return the finished string
        return Line(line, f_file, f_line)



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import LineDirective
from .common import directive
from ..common import PrologueError

@directive("include")
class Include(LineDirective):
    """ Includes the body of another file into the output stream """

    def __init__(self, parent, src_file=None, src_line=0, callback=None):
        super().__init__(
            parent, yields=True, src_file=src_file, src_line=src_line,
            callback=callback,
        )
        self.filename = None

    def invoke(self, tag, arguments):
        """ Include a different file into the stream.

        Args:
            tag      : Tag used to trigger this directive
            arguments: Argument string provided to the directive
        """
        # Sanity check
        if tag != "include":
            raise PrologueError(f"Include invoked with '{tag}'")
        if self.count_args(arguments) != 1:
            raise PrologueError(f"Invalid form used for #include {arguments}")
        # Store the filename
        self.filename = self.get_arg(arguments, 0)

    def evaluate(self, context):
        """ Include a file.

        Args:
            context: The content object at the point of evaluation.

        Yields: Stream of lines from the included file
        """
        yield from context.pro.evaluate_inner(self.filename, context=context)

@directive("import")
class Import(LineDirective):
    """ Includes a file, but only once per preprocessing session """

    def __init__(self, parent, src_file=None, src_line=0, callback=None):
        super().__init__(
            parent, yields=True, src_file=src_file, src_line=src_line,
            callback=callback,
        )
        self.filename = None

    def invoke(self, tag, arguments):
        """ Include a different file into the stream.

        Args:
            tag      : Tag used to trigger this directive
            arguments: Argument string provided to the directive
        """
        # Sanity check
        if tag != "import":
            raise PrologueError(f"Import invoked with '{tag}'")
        if self.count_args(arguments) != 1:
            raise PrologueError(f"Invalid form used for #import {arguments}")
        # Store the filename
        self.filename = self.get_arg(arguments, 0)

    def evaluate(self, context):
        """ Import a file (include only once).

        Args:
            context: The content object at the point of evaluation.

        Yields: Stream of lines from the imported file
        """
        r_file = context.pro.registry.resolve(self.filename)
        if r_file not in context.trace:
            yield from context.pro.evaluate_inner(self.filename, context=context)



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTE: 'return' then 'yield' is used to avoid yielding 'None'

from .base import LineDirective
from .common import directive
from ..common import PrologueError

INFO    = ["info"]
WARNING = ["warn", "warning", "todo", "fixme"]
ERROR   = ["error", "danger", "fatal"]

@directive(*INFO, *WARNING, *ERROR)
class Message(LineDirective):
    """ Prints message at different verbosity levels """

    def __init__(self, parent, src_file=None, src_line=0, callback=None):
        super().__init__(
            parent, src_file=src_file, src_line=src_line, callback=callback,
        )
        self.msg_class = None
        self.msg_text  = None

    def invoke(self, tag, arguments):
        """ Trigger a message to be printed.

        Args:
            context  : The current context object
            tag      : Tag used to trigger this directive
            arguments: Argument string provided to the directive
        """
        global INFO, WARNING, ERROR
        self.msg_text = arguments
        if   tag in INFO   : self.msg_class = "INFO"
        elif tag in WARNING: self.msg_class = "WARNING"
        elif tag in ERROR  : self.msg_class = "ERROR"
        else               : raise PrologueError(f"Unrecognised message type {tag}")

    def evaluate(self, context):
        """ Print the message.

        Args:
            context: The context object at the point of evaluation.
        """
        {
            "INFO"   : context.pro.info_message,
            "WARNING": context.pro.warning_message,
            "ERROR"  : context.pro.error_message,
        }[self.msg_class](self.msg_text, source=self.source)



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .conditional import Conditional
from .define import Define, Undefine
from .for_loop import ForLoop
from .include import Include, Import
from .message import Message

def register_prime_directives(pro):
    """ Register prime directives onto an instance of Prologue.

    Args:
        pro: Pointer to Prologue instance
    """
    # Register message directives
    pro.register_directive(Message)
    # Variable define/undefine
    pro.register_directive(Define)
    pro.register_directive(Undefine)
    # Register logical directives
    pro.register_directive(Conditional)
    pro.register_directive(ForLoop)
    # Include/import directives
    pro.register_directive(Include)
    pro.register_directive(Import)



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from setuptools import setup, find_packages

with open(Path(__file__).parent / "README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="prologue",
    version="1.1",
    license="Apache License, Version 2.0",
    description="General purpose preprocessor written in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Peter Birch",
    url="https://github.com/Intuity",
    project_urls={
        "Source": "https://github.com/Intuity/prologue",
    },
    packages=find_packages(exclude=["tests"]),
    data_files=["LICENSE"],
    include_package_data=True,
    entry_points={ },
    python_requires=">=3.6.10",
    install_requires=["asteval", "astunparse"],
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "pytest-cov", "pytest-mock"],
    extras_require={},
)



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shlex

from ..common import PrologueError
from ..block import Block

class Directive(Block):
    UUID = 0

    def __init__(
        self, parent, yields=True, src_file=None, src_line=0, callback=None,
    ):
        """ Initialise the directive.

        Args:
            parent  : The parent block
            yields  : Whether the directive yields content
            src_file: Source file
            src_line: Source line number
            callback: External callback routine to expose parse state
        """
        super().__init__(parent)
        self.__uuid     = BlockDirective.issue_uuid()
        self.__yields   = yields
        self.__source   = (src_file, src_line)
        self.__callback = callback

    @property
    def yields(self): return self.__yields

    @property
    def uuid(self): return self.__uuid

    @property
    def source(self): return self.__source

    @property
    def callback(self): return self.__callback

    @classmethod
    def issue_uuid(self):
        issue = Directive.UUID
        Directive.UUID += 1
        return issue

    def split_args(self, args):
        """ Split an argument string into parts, paying attention to quotes.

        Args:
            args: The string to split

        Returns: Array of sections
        """
        return shlex.split(args)

    def count_args(self, args):
        """ Count the number of arguments found, paying attention to quotes.

        Args:
            args: The string to split

        Returns: Number of arguments found
        """
        return len(self.split_args(args))

    def get_arg(self, args, index, default=None):
        """ Get the Nth argument from a string paying attention to quotes.

        Args:
            args   : The string to parse
            index  : The section to extract
            default: The default value to return if index out of range
        """
        parts = self.split_args(args)
        return parts[index] if index < len(parts) else default

class BlockDirective(Directive):
    """
    A block directive contains lines and other directives for which evaluation
    is postponed until the block is closed. Blocks must have an opening and
    closing tag, but can also be split into multiple sections using transitions.
    """

    def __init__(
        self, parent, yields=True, src_file=None, src_line=0, callback=None
    ):
        """ Initialise the block directive.

        Args:
            parent  : The parent block
            yields  : Whether the directive yields content (defaults to True for block)
            src_file: Source file
            src_line: Source line number
            callback: External callback routine to expose parse state
        """
        super().__init__(parent, yields, src_file, src_line, callback)
        self.__opened = False
        self.__closed = False

    @property
    def opened(self): return self.__opened

    @property
    def closed(self): return self.__closed

    def open(self, tag, arguments):
        """ Called once for the opening directive of the block.

        Args:
            tag      : The tag that opens the directive
            arguments: Argument string following the directive
        """
        if self.__opened:
            raise PrologueError("Multiple opening statements for block detected")
        self.__opened = True

    def transition(self, tag, arguments):
        """ Called for every transition between different sections of the directive.

        Args:
            tag      : The tag that opens the section
            arguments: Argument string following the directive
        """
        if not self.__opened:
            raise PrologueError(f"Transition '{tag}' used before opening directive")
        elif self.__closed:
            raise PrologueError(f"Transition '{tag}' used after closing directive")

    def close(self, tag, arguments):
        """ Called once to close the directive.

        Args:
            tag      : The tag that opens the directive
            arguments: Argument string following the directive
        """
        if self.__closed:
            raise PrologueError("Multiple closing statements for block detected")
        self.__closed = True

    def append(self, entry):
        """ Append a new line or nested block to the active block.

        Args:
            entry: The line or block to append.
        """
        # Sanity checks
        if not self.opened:
            raise PrologueError("Trying to append a line to an unopened directive")
        elif self.closed:
            raise PrologueError("Trying to append a line to a closed directive")
        # Call append on parent class
        super().append(entry)

class LineDirective(Directive):
    """
    A line directive performs a single action. Evaluation is only postponed if
    it is embedded within a block directive.
    """

    def __init__(
        self, parent, yields=False, src_file=None, src_line=0, callback=None,
    ):
        """ Initialise the block directive.

        Args:
            parent  : The parent block
            yields  : Whether the directive yields content (defaults to False for line)
            src_file: Source file
            src_line: Source line number
            callback: External callback routine to expose parse state
        """
        super().__init__(parent, yields, src_file, src_line, callback)

    def invoke(self, tag, arguments):
        """ Called once to setup the directive.

        Args:
            tag      : The tag that opens the directive
            arguments: Argument string following the directive
        """
        raise PrologueError("Must provide implementation of 'invoke'")



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import shlex

import asteval

from .base import BlockDirective
from ..block import Block
from .common import directive
from ..common import PrologueError

@directive(opening=["for"], closing=["endfor"])
class ForLoop(BlockDirective):
    """ Loop directive using FOR/ENDFOR """

    def __init__(self, parent, src_file=None, src_line=0, callback=None):
        super().__init__(
            parent, src_file=src_file, src_line=src_line, callback=callback,
        )
        self.loop = None

    def open(self, tag, arguments):
        # Sanity checks
        if tag != "for":
            raise PrologueError(f"Loop opening invoked with '{tag}'")
        # Only open loop after the sanity checks
        super().open(tag, arguments)
        # Record the loop argument
        self.loop = arguments

    def transition(self, tag, arguments):
        raise PrologueError("For loop does not support transitions")

    def close(self, tag, arguments):
        # Sanity checks
        if tag != "endfor":
            raise PrologueError(f"Loop close invoked with '{tag}'")
        # Only close loop after the sanity checks
        super().close(tag, arguments)

    RGX_RANGE = re.compile(r"^range[\s]*[(][\s]*(.*?)[\s]*[)]$")
    RGX_ARRAY = re.compile(r"^[\[](.*?)[\]]$")

    def evaluate(self, context):
        """ Repeats the embedded block as many times as required.

        Args:
            context: The context object at the point of evaluation

        Yields: A line of text at a time
        """
        # Locate the 'in' keyword
        parts = shlex.split(self.loop)
        if "in" not in parts:
            raise PrologueError(f"Incorrectly formed loop condition '{self.loop}'")
        pre_loop  = (" ".join(parts[:parts.index("in")])).strip()
        post_loop = (" ".join(parts[parts.index("in")+1:])).strip()
        # Recognise a range
        m_range  = self.RGX_RANGE.match(post_loop)
        m_array  = self.RGX_ARRAY.match(post_loop)
        iterable = None
        if m_range:
            iterable = range(context.evaluate(m_range.groups(0)[0]))
        elif m_array:
            iterable = [x.strip() for x in m_array.groups(0)[0].split(",")]
        else:
            iterable = asteval.Interpreter()(context.flatten(post_loop))
        # TODO: Need to support complex unpacking of loop conditions
        for entry in iterable:
            # Expose the current value of the loop variable
            context.set_define(pre_loop, entry, check=False)
            # Perform substitutions for the loop variable (and any others)
            for line in super().evaluate(context):
                yield context.substitute(line)



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntEnum, auto

from .base import Directive, BlockDirective, LineDirective
from ..common import PrologueError

class DirectiveWrap(object):
    """ Decorator around a directive class or function """

    def __init__(self, dirx, opening, closing=None, transition=None):
        """ Initialise the wrapper

        Args:
            dirx      : The directive class or function
            dir_type  : Type of the directive (e.g. BLOCK or LINE)
            opening   : List of opening tags for the directive
            closing   : List of closing tags for the directive (only used for block)
            transition: Optional list of tags which transition between sections
        """
        # Store and check directive
        self.directive = dirx
        if not issubclass(dirx, Directive):
            raise PrologueError(f"Item is not a subclass of Directive: {dirx}")
        # Force tags to be case insensitive
        if opening   : opening    = [x.lower() for x in opening   ]
        if closing   : closing    = [x.lower() for x in closing   ]
        if transition: transition = [x.lower() for x in transition]
        # Check tags
        opening    = tuple(opening) if isinstance(opening, list) else opening
        closing    = tuple(closing) if isinstance(closing, list) else closing
        transition = tuple(transition) if isinstance(transition, list) else transition
        if not isinstance(opening, tuple) or len(opening) == 0:
            raise PrologueError("At least one opening tag must be specified")
        if self.is_block and (not isinstance(closing, tuple) or len(closing) == 0):
            raise PrologueError("At least one closing tag must be specified")
        elif not self.is_block and closing != None:
            raise PrologueError("Only a block directive can have closing tags")
        elif not self.is_block and transition != None:
            raise PrologueError("Only a block directive can have transition tags")
        for tag in (
            opening +
            (closing if closing else tuple()) +
            (transition if transition else tuple())
        ):
            if " " in tag or len(tag) == 0:
                raise PrologueError(
                    "Directive tag cannot contain spaces and must be at least "
                    "one character in length"
                )
        # Store arguments
        self.opening    = opening
        self.closing    = closing if closing else tuple()
        self.transition = transition if transition else tuple()

    @property
    def tags(self):
        """ Return all tags - opening and closing """
        return (self.opening + self.transition + self.closing)

    @property
    def is_block(self):
        """ Returns if this is a block directive """
        return (
            isinstance(self.directive, type) and
            issubclass(self.directive, BlockDirective)
        )

    @property
    def is_line(self):
        """ Returns if this is a line directive """
        return (
            isinstance(self.directive, type) and
            issubclass(self.directive, LineDirective)
        )

    def is_opening(self, tag):
        """ Test if a particular tag is an opening tag.

        Args:
            tag: The tag to test

        Returns: True if this is an opening tag, False otherwise
        """
        if   tag in self.opening: return True
        elif tag in (self.closing + self.transition): return False
        else: raise PrologueError(f"Tag is not known by directive: {tag}")

    def is_transition(self, tag):
        """ Test if a particular tag is a transition tag.

        Args:
            tag: The tag to test

        Returns: True if this is a transition tag, False otherwise
        """
        if   tag in self.transition: return True
        elif tag in (self.opening + self.closing): return False
        else: raise PrologueError(f"Tag is not known by directive: {tag}")

    def is_closing(self, tag):
        """ Test if a particular tag is a closing tag.

        Args:
            tag: The tag to test

        Returns: True if this is a closing tag, False otherwise
        """
        if   tag in self.closing: return True
        elif tag in (self.opening + self.transition): return False
        else: raise PrologueError(f"Tag is not known by directive: {tag}")

def directive(*tags, opening=None, closing=None, transition=None):
    """ Decorator for a block directive

    Args:
        *tags     : List of tags identifying directive (alias of opening)
        opening   : Tags that open a block directive
        closing   : Tags that close a block directive
        transition: Tags that transition between sections of the block
    """
    # Pickup tags list if explicit opening list not provided
    if not opening: opening = []
    opening += tags
    # Force tags to lowercase
    opening = [x.lower() for x in opening]
    if transition: transition = [x.lower() for x in transition]
    if closing   : closing    = [x.lower() for x in closing   ]
    # Setup the wrapper function
    def wrapper(_dirx):
        if not opening:
            raise PrologueError(
                "At least one opening directive must be given"
            )
        if not issubclass(_dirx, BlockDirective) and not issubclass(_dirx, LineDirective):
            raise PrologueError(
                "Directive must be a subclass of BlockDirective or LineDirective"
            )
        elif issubclass(_dirx, LineDirective) and (closing or transition):
            raise PrologueError(
                "Closing or transition tags can only be used with a block directive"
            )
        elif issubclass(_dirx, BlockDirective) and not closing:
            raise PrologueError(
                "Closing tags must be specified for a block directive"
            )
        # Record tags against the registered class
        _dirx.OPENING    = opening
        _dirx.TRANSITION = transition
        _dirx.CLOSING    = closing
        # Return the wrapped directive
        return DirectiveWrap(_dirx, opening, closing, transition)
    return wrapper



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import LineDirective
from .common import directive
from ..common import PrologueError

@directive("define")
class Define(LineDirective):
    """ Defines a variable in the current context """

    def __init__(self, parent, src_file=None, src_line=0, callback=None):
        super().__init__(
            parent, src_file=src_file, src_line=src_line, callback=callback,
        )
        self.name  = None
        self.value = None

    def invoke(self, tag, arguments):
        """ Define a new variable in the context object.

        Args:
            tag      : Tag used to trigger this directive
            arguments: Argument string provided to the directive
        """
        # Sanity checks
        if tag != "define":
            raise PrologueError(f"Define invoked with '{tag}'")
        # Test number of arguments
        # NOTE: At least one argument (the name) must be provided, but as an
        #       expression can be used in a define, there is no upper limit on
        #       arguments as counted by shlex
        num_args = self.count_args(arguments)
        if num_args < 1: raise PrologueError("Invalid form used for #define")
        # Everything before the first space is taken as the variable name, and
        # everything after the first space is the value
        self.name  = self.get_arg(arguments, 0)
        self.value = arguments[len(self.name)+1:].strip()
        # If no value is provided, it acts like a flag with value 'True'
        if len(self.value) == 0: self.value = True

    def evaluate(self, context):
        """ Define a variable.

        Args:
            context: The context object at the point of evaluation
        """
        context.set_define(self.name, self.value, check=True)

@directive("undef")
class Undefine(LineDirective):
    """ Undefined a variable from the current context """

    def __init__(self, parent, src_file=None, src_line=0, callback=None):
        super().__init__(
            parent, src_file=src_file, src_line=src_line, callback=callback,
        )
        self.name = None

    def invoke(self, tag, arguments):
        """ Undefine an existing variable from the context object.

        Args:
            tag      : Tag used to trigger this directive
            arguments: Argument string provided to the directive
        """
        # Sanity check
        if tag != "undef":
            raise PrologueError(f"Undefine invoked with '{tag}'")
        if self.count_args(arguments) != 1:
            raise PrologueError(f"Invalid form used for #undef {arguments}")
        # Everything before the first space is taken as the variable name
        self.name = self.get_arg(arguments, 0)

    def evaluate(self, context):
        """ Undefine a variable.

        Args:
            context: The context object at the point of evaluation
        """
        context.clear_define(self.name)



# Copyright 2020, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .common import PrologueError

class Block:
    """ Represents a block of lines between opening and closing delimiters """

    def __init__(self, parent):
        """ Initialise a block with a directive. """
        self.parent  = parent
        self.content = []

    def append(self, entry):
        """ Append either a new line of text or a nested Block instance.

        Args:
            entry: The line of text or nested Block
        """
        if not isinstance(entry, str) and not isinstance(entry, Block):
            raise PrologueError(
                f"Entry must be a string or Block, not {type(entry).__name__}"
            )
        self.content.append(entry)

    def evaluate(self, context):
        """ Evaluate the block and stream complete lines back to Prologue.

        Args:
            context: Context object at the point of evaluation

        Yields: A line of text at a time
        """
        from .directives.base import Directive
        for entry in self.content:
            if (
                (    isinstance(entry, Directive) and entry.yields            ) or
                (not isinstance(entry, Directive) and isinstance(entry, Block))
            ):
                yield from entry.evaluate(context)
            elif isinstance(entry, Directive):
                entry.evaluate(context)
            else:
                yield entry

    @property
    def stack(self):
        """ Returns the full stack of nodes including this one. """
        baseline = self.parent.stack if self.parent else []
        return baseline + [self]
